@@ -25,7 +25,7 @@ from typing import Union
 from typing import overload
 from warnings import warn
 
-import toml as _toml
+import yaml
 from requests.adapters import HTTPAdapter
 from requests.adapters import MaxRetryError
 from requests.exceptions import ConnectionError
@@ -230,6 +230,14 @@ class CallList(Sequence[Any], Sized):
     def __len__(self) -> int:
         return len(self._calls)
 
+    @overload
+    def __getitem__(self, idx: int) -> Call:
+        """Overload when get a single item."""
+
+    @overload
+    def __getitem__(self, idx: slice) -> List[Call]:
+        """Overload when a slice is requested."""
+
     def __getitem__(self, idx: Union[int, slice]) -> Union[Call, List[Call]]:
         return self._calls[idx]
 
@@ -556,6 +564,7 @@ class Response(BaseResponse):
 
     def get_response(self, request: "PreparedRequest") -> HTTPResponse:
         if self.body and isinstance(self.body, Exception):
+            setattr(self.body, "request", request)
             raise self.body
 
         headers = self.get_headers()
@@ -660,7 +669,7 @@ class RequestsMock(object):
         passthru_prefixes: Tuple[str, ...] = (),
         target: str = "requests.adapters.HTTPAdapter.send",
         registry: Type[FirstMatchRegistry] = FirstMatchRegistry,
-    ):
+    ) -> None:
         self._calls: CallList = CallList()
         self.reset()
         self._registry: FirstMatchRegistry = registry()  # call only after reset
@@ -769,9 +778,15 @@ class RequestsMock(object):
     post = partialmethod(add, POST)
     put = partialmethod(add, PUT)
 
+    def _parse_response_file(
+        self, file_path: "Union[str, bytes, os.PathLike[Any]]"
+    ) -> "Dict[str, Any]":
+        with open(file_path, "r") as file:
+            data = yaml.safe_load(file)
+        return data
+
     def _add_from_file(self, file_path: "Union[str, bytes, os.PathLike[Any]]") -> None:
-        with open(file_path) as file:
-            data = _toml.load(file)
+        data = self._parse_response_file(file_path)
 
         for rsp in data["responses"]:
             rsp = rsp["response"]
@@ -910,14 +925,15 @@ class RequestsMock(object):
 
     def __exit__(self, type: Any, value: Any, traceback: Any) -> bool:
         success = type is None
-        self.stop(allow_assert=success)
-        self.reset()
+        try:
+            self.stop(allow_assert=success)
+        finally:
+            self.reset()
         return success
 
     @overload
     def activate(self, func: _F = ...) -> _F:
         """Overload for scenario when 'responses.activate' is used."""
-        ...  # pragma: no cover
 
     @overload
     def activate(
@@ -930,9 +946,7 @@ class RequestsMock(object):
         'responses.activate(registry=, assert_all_requests_are_fired=True)' is used.
 
         See https://github.com/getsentry/responses/pull/469 for more details
-
         """
-        ...  # pragma: no cover
 
     def activate(
         self,
