@@ -13,10 +13,10 @@ from unittest.mock import patch
 
 import pytest
 import requests
-from requests.adapters import MaxRetryError
 from requests.exceptions import ChunkedEncodingError
 from requests.exceptions import ConnectionError
 from requests.exceptions import HTTPError
+from requests.exceptions import RetryError
 from urllib3.util.retry import Retry
 
 import responses
@@ -1063,10 +1063,10 @@ def test_response_filebody():
     def run():
         current_file = os.path.abspath(__file__)
         with responses.RequestsMock() as m:
-            with open(current_file, "r") as out:
+            with open(current_file, "r", encoding="utf-8") as out:
                 m.add(responses.GET, "http://example.com", body=out.read(), stream=True)
                 resp = requests.get("http://example.com", stream=True)
-            with open(current_file, "r") as out:
+            with open(current_file, "r", encoding="utf-8") as out:
                 assert resp.text == out.read()
 
     run()
@@ -1133,6 +1133,28 @@ def test_assert_all_requests_are_fired():
     assert_reset()
 
 
+def test_assert_all_requests_fired_multiple():
+    @responses.activate(assert_all_requests_are_fired=True)
+    def test_some_function():
+        # Not all mocks are called so we'll get an AssertionError
+        responses.add(responses.GET, "http://other_url", json={})
+        responses.add(responses.GET, "http://some_api", json={})
+        requests.get("http://some_api")
+
+    @responses.activate(assert_all_requests_are_fired=True)
+    def test_some_second_function():
+        # This should pass as mocks should be reset.
+        responses.add(responses.GET, "http://some_api", json={})
+        requests.get("http://some_api")
+
+    with pytest.raises(AssertionError):
+        test_some_function()
+    assert_reset()
+
+    test_some_second_function()
+    assert_reset()
+
+
 def test_allow_redirects_samehost():
     redirecting_url = "http://example.com"
     final_url_path = "/1"
@@ -1566,19 +1588,21 @@ def test_multiple_methods():
 
 
 def test_passthrough_flag(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
-    response = Response(responses.GET, httpserver.url, body="MOCK")
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+    url = httpserver.url_for("/")
+
+    response = Response(responses.GET, url, body="MOCK")
 
     @responses.activate
     def run_passthrough():
         responses.add(response)
-        resp = requests.get(httpserver.url)
+        resp = requests.get(url)
         assert_response(resp, "OK")
 
     @responses.activate
     def run_mocked():
         responses.add(response)
-        resp = requests.get(httpserver.url)
+        resp = requests.get(url)
         assert_response(resp, "MOCK")
 
     run_mocked()
@@ -1589,37 +1613,65 @@ def test_passthrough_flag(httpserver):
     assert_reset()
 
 
+def test_passthrough_kwarg(httpserver):
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+    url = httpserver.url_for("/")
+
+    def configure_response(passthrough):
+        responses.get(url, body="MOCK", passthrough=passthrough)
+
+    @responses.activate
+    def run_passthrough():
+        configure_response(passthrough=True)
+        resp = requests.get(url)
+        assert_response(resp, "OK")
+
+    @responses.activate
+    def run_mocked():
+        configure_response(passthrough=False)
+        resp = requests.get(url)
+        assert_response(resp, "MOCK")
+
+    run_mocked()
+    assert_reset()
+
+    run_passthrough()
+    assert_reset()
+
+
 def test_passthrough_response(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+    url = httpserver.url_for("/")
 
     @responses.activate
     def run():
-        responses.add(PassthroughResponse(responses.GET, httpserver.url))
-        responses.add(responses.GET, "{}/one".format(httpserver.url), body="one")
+        responses.add(PassthroughResponse(responses.GET, url))
+        responses.add(responses.GET, "{}/one".format(url), body="one")
         responses.add(responses.GET, "http://example.com/two", body="two")
 
         resp = requests.get("http://example.com/two")
         assert_response(resp, "two")
-        resp = requests.get("{}/one".format(httpserver.url))
+        resp = requests.get("{}/one".format(url))
         assert_response(resp, "one")
-        resp = requests.get(httpserver.url)
+        resp = requests.get(url)
         assert_response(resp, "OK")
 
         assert len(responses.calls) == 3
-        responses.assert_call_count(httpserver.url, 1)
+        responses.assert_call_count(url, 1)
 
     run()
     assert_reset()
 
 
 def test_passthrough_response_stream(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
 
     @responses.activate
     def run():
-        responses.add(PassthroughResponse(responses.GET, httpserver.url))
-        content_1 = requests.get(httpserver.url).content
-        with requests.get(httpserver.url, stream=True) as resp:
+        url = httpserver.url_for("/")
+        responses.add(PassthroughResponse(responses.GET, url))
+        content_1 = requests.get(url).content
+        with requests.get(url, stream=True) as resp:
             content_2 = resp.raw.read()
         assert content_1 == content_2
 
@@ -1628,19 +1680,20 @@ def test_passthrough_response_stream(httpserver):
 
 
 def test_passthru_prefixes(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+    url = httpserver.url_for("/")
 
     @responses.activate
     def run_constructor_argument():
-        with responses.RequestsMock(passthru_prefixes=(httpserver.url,)):
-            resp = requests.get(httpserver.url)
+        with responses.RequestsMock(passthru_prefixes=(url,)):
+            resp = requests.get(url)
             assert_response(resp, "OK")
 
     @responses.activate
     def run_property_setter():
         with responses.RequestsMock() as m:
-            m.passthru_prefixes = tuple([httpserver.url])
-            resp = requests.get(httpserver.url)
+            m.passthru_prefixes = tuple([url])
+            resp = requests.get(url)
             assert_response(resp, "OK")
 
     run_constructor_argument()
@@ -1650,19 +1703,20 @@ def test_passthru_prefixes(httpserver):
 
 
 def test_passthru(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+    url = httpserver.url_for("/")
 
     @responses.activate
     def run():
-        responses.add_passthru(httpserver.url)
-        responses.add(responses.GET, "{}/one".format(httpserver.url), body="one")
+        responses.add_passthru(url)
+        responses.add(responses.GET, "{}/one".format(url), body="one")
         responses.add(responses.GET, "http://example.com/two", body="two")
 
         resp = requests.get("http://example.com/two")
         assert_response(resp, "two")
-        resp = requests.get("{}/one".format(httpserver.url))
+        resp = requests.get("{}/one".format(url))
         assert_response(resp, "one")
-        resp = requests.get(httpserver.url)
+        resp = requests.get(url)
         assert_response(resp, "OK")
 
     run()
@@ -1670,21 +1724,24 @@ def test_passthru(httpserver):
 
 
 def test_passthru_regex(httpserver):
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request(re.compile("^/\\w+")).respond_with_data(
+        "OK", content_type="text/plain"
+    )
+    url = httpserver.url_for("/")
 
     @responses.activate
     def run():
-        responses.add_passthru(re.compile("{}/\\w+".format(httpserver.url)))
-        responses.add(responses.GET, "{}/one".format(httpserver.url), body="one")
+        responses.add_passthru(re.compile(f"{url}/\\w+"))
+        responses.add(responses.GET, "{}/one".format(url), body="one")
         responses.add(responses.GET, "http://example.com/two", body="two")
 
         resp = requests.get("http://example.com/two")
         assert_response(resp, "two")
-        resp = requests.get("{}/one".format(httpserver.url))
+        resp = requests.get(f"{url}/one")
         assert_response(resp, "one")
-        resp = requests.get("{}/two".format(httpserver.url))
+        resp = requests.get(f"{url}/two")
         assert_response(resp, "OK")
-        resp = requests.get("{}/three".format(httpserver.url))
+        resp = requests.get(f"{url}/three")
         assert_response(resp, "OK")
 
     run()
@@ -1697,7 +1754,7 @@ def test_passthru_does_not_persist_across_tests(httpserver):
     see:
     https://github.com/getsentry/responses/issues/322
     """
-    httpserver.serve_content("OK", headers={"Content-Type": "text/plain"})
+    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
 
     @responses.activate
     def with_a_passthru():
@@ -2156,9 +2213,12 @@ class TestMultipleWrappers:
             Mock real HTTP server
 
         """
-        httpserver.serve_content("OK", code=969, headers={"Content-Type": "text/plain"})
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain", status=969
+        )
+        url = httpserver.url_for("/")
 
-        response = requests.get(httpserver.url)
+        response = requests.get(url)
         assert response.status_code == 969
 
 
@@ -2348,6 +2408,21 @@ def test_redirect():
 
 
 class TestMaxRetry:
+    def set_session(self, total=4, raise_on_status=True):
+        session = requests.Session()
+
+        adapter = requests.adapters.HTTPAdapter(
+            max_retries=Retry(
+                total=total,
+                backoff_factor=0.1,
+                status_forcelist=[500],
+                method_whitelist=["GET", "POST", "PATCH"],
+                raise_on_status=raise_on_status,
+            )
+        )
+        session.mount("https://", adapter)
+        return session
+
     def test_max_retries(self):
         """This example is present in README.rst"""
 
@@ -2359,17 +2434,7 @@ class TestMaxRetry:
             rsp3 = responses.get(url, body="Error", status=500)
             rsp4 = responses.get(url, body="OK", status=200)
 
-            session = requests.Session()
-
-            adapter = requests.adapters.HTTPAdapter(
-                max_retries=Retry(
-                    total=4,
-                    backoff_factor=0.1,
-                    status_forcelist=[500],
-                    method_whitelist=["GET", "POST", "PATCH"],
-                )
-            )
-            session.mount("https://", adapter)
+            session = self.set_session()
 
             resp = session.get(url)
 
@@ -2391,21 +2456,10 @@ class TestMaxRetry:
             rsp2 = responses.get(url, body="Error", status=500)
             rsp3 = responses.get(url, body="Error", status=500)
 
-            session = requests.Session()
-
-            adapter = requests.adapters.HTTPAdapter(
-                max_retries=Retry(
-                    total=2,
-                    backoff_factor=0.1,
-                    status_forcelist=[500],
-                    method_whitelist=["GET", "POST", "PATCH"],
-                    raise_on_status=raise_on_status,
-                )
-            )
-            session.mount("https://", adapter)
+            session = self.set_session(total=2, raise_on_status=raise_on_status)
 
             if raise_on_status:
-                with pytest.raises(MaxRetryError):
+                with pytest.raises(RetryError):
                     session.get(url)
             else:
                 resp = session.get(url)
@@ -2418,6 +2472,23 @@ class TestMaxRetry:
         run()
         assert_reset()
 
+    def test_max_retries_exceed_msg(self):
+        @responses.activate(registry=registries.OrderedRegistry)
+        def run():
+            url = "https://example.com"
+            responses.get(url, body="Error", status=500)
+            responses.get(url, body="Error", status=500)
+
+            session = self.set_session(total=1)
+
+            with pytest.raises(RetryError) as err:
+                session.get(url)
+
+            assert "too many 500 error responses" in str(err.value)
+
+        run()
+        assert_reset()
+
     def test_adapter_retry_untouched(self):
         """Validate that every new request uses brand-new Retry object"""
 
@@ -2434,18 +2505,7 @@ class TestMaxRetry:
             responses.add(error_rsp)
             responses.add(ok_rsp)
 
-            session = requests.Session()
-
-            adapter = requests.adapters.HTTPAdapter(
-                max_retries=Retry(
-                    total=4,
-                    backoff_factor=0.1,
-                    status_forcelist=[500],
-                    method_whitelist=["GET", "POST", "PATCH"],
-                )
-            )
-            session.mount("https://", adapter)
-
+            session = self.set_session()
             resp = session.get(url)
             assert resp.status_code == 200
 
@@ -2456,3 +2516,20 @@ class TestMaxRetry:
 
         run()
         assert_reset()
+
+
+def test_request_object_attached_to_exception():
+    """Validate that we attach `request` object to custom exception supplied as body"""
+
+    @responses.activate
+    def run():
+        url = "https://httpbin.org/delay/2"
+        responses.get(url, body=requests.ReadTimeout())
+
+        try:
+            requests.get(url, timeout=1)
+        except requests.ReadTimeout as exc:
+            assert type(exc.request) == requests.models.PreparedRequest
+
+    run()
+    assert_reset()
